{% if vulns %}
<figure>
<table class="sortable">
    <thead>
        <tr><th>Severity</th><th>Name</th><th>Source</th><th>Location</th><th>State</th></tr>
    </thead>
    <tbody>
        {% for v in vulns %}
        <tr>
            <td><span class="badge-{{ v.severity }}">{{ v.severity }}</span></td>
            <td><a href="{{ v.web_url }}" target="_blank">{{ v.name }}</a></td>
            <td>{{ v.source }}</td>
            <td>{{ v.location or "" }}</td>
            <td>{{ v.state }}</td>
        </tr>
        {% endfor %}
    </tbody>
</table>
</figure>
{% else %}
<p>No vulnerabilities.</p>
{% endif %}
